import jug.task
import jug.dict_store


jug.task.Task.store = jug.dict_store.dict_store()

def add1(x):
    return x + 1
def add2(x):
    return x + 2

def _assert_tsorted(tasks):
    for i in xrange(len(tasks)):
        for j in xrange(i+1,len(tasks)):
            for dep in list(tasks[i].dependencies) + tasks[i].kwdependencies.values():
                if type(dep) is list:
                    assert tasks[j] not in dep
                else:
                    assert tasks[j] is not dep
            
def test_topological_sort():
    bases = [jug.task.Task(add1,i) for i in xrange(10)]
    derived = [jug.task.Task(add1,t) for t in bases]
    derived2 = [jug.task.Task(add1,t) for t in derived]
    derived3 = [jug.task.Task(add1,t) for t in derived]
    
    alltasks = bases + derived
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived)
    _assert_tsorted(alltasks)
    
    alltasks.reverse()
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived)
    _assert_tsorted(alltasks)

    alltasks = bases + derived
    alltasks.reverse()
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived)
    _assert_tsorted(alltasks)

    alltasks = derived + bases
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived)
    _assert_tsorted(alltasks)

    alltasks = derived + bases + derived2
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived) + len(derived2)
    _assert_tsorted(alltasks)

    alltasks = derived + bases + derived2 + derived3
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived) + len(derived2) + len(derived3)
    _assert_tsorted(alltasks)

def test_topological_sort_kwargs():
    def add2(x):
        return x + 2
    def sumlst(lst,param):
        return sum(lst)

    bases = [jug.task.Task(add2,x=i) for i in xrange(10)]
    derived = [jug.task.Task(sumlst,lst=bases,param=p) for p in xrange(4)]

    alltasks = bases + derived
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived)
    _assert_tsorted(alltasks)
    
    alltasks.reverse()
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived)
    _assert_tsorted(alltasks)

    alltasks = bases + derived
    alltasks.reverse()
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived)
    _assert_tsorted(alltasks)

    alltasks = derived + bases
    jug.task.topological_sort(alltasks)
    assert len(alltasks) == len(bases) + len(derived)
    _assert_tsorted(alltasks)


def test_load_after_run():
    T = jug.task.Task(add1,1)
    T.run()
    assert T.can_load()

def test_hash_same_func():
    T0 = jug.task.Task(add1,0)
    T1 = jug.task.Task(add1,1)

    assert T0.hash() != T1.hash()
    
def test_hash_different_func():
    T0 = jug.task.Task(add1,0)
    T1 = jug.task.Task(add2,0)

    assert T0.hash() != T1.hash()


def test_taskgenerator():
    @jug.task.TaskGenerator
    def double(x):
        return 2*x
    T=double(2)
    assert type(T) == jug.task.Task
    assert not T.print_result
    @jug.task.TaskGenerator(print_result=True)
    def square(x):
        return x*x
    T=square(2)
    assert type(T) == jug.task.Task
    assert T.print_result


def test_unload():
    T0 = jug.task.Task(add1,0)
    assert not T0.finished
    assert T0.can_run()
    T0.run()
    assert T0.finished
    assert T0.result == 1
    T0.unload()
    assert T0.result == 1

def identity(x):
    return x

def test_cachedfunction():
    assert jug.task.CachedFunction(identity,123) == 123
    assert jug.task.CachedFunction(identity,'mixture') == 'mixture'

def test_npyload():
    import numpy as np
    A = np.arange(23)
    assert np.all(jug.task.CachedFunction(identity,A) == A)



# -*- coding: utf-8 -*-
# Copyright (C) 2008-2009, Luís Pedro Coelho <lpc@cmu.edu>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

'''
file_store : file-system based data store & locks.
'''
from __future__ import division

try:
    import cPickle as pickle
except ImportError:
    import pickle
import os
from os import path, mkdir, fdopen
from os.path import dirname, exists
import errno
import tempfile
import options
import numpy as np
from gzip import GzipFile

def _fsize(fname):
    '''Returns file size'''
    return os.stat_result(os.stat(fname)).st_size


def create_directories(dname):
    '''
    create_directories(dname)

    Recursively create directories.
    '''
    if dname.endswith('/'): dname = dname[:-1]
    head, tail = path.split(dname)
    if head: create_directories(head)
    try:
        mkdir(dname)
    except OSError, e:
        if e.errno != errno.EEXIST:
            raise


class file_store(object):
    def __init__(self, dname):
        '''
        file_store(dname)

        Recursively create directories.
        '''
        if dname.endswith('/'): dname = dname[:-1]
        self.jugdir = dname
        head, tail = path.split(dname)
        if head: create_directories(head)
        create_directories(options.tempdir)
        try:
            mkdir(dname)
        except OSError, e:
            if e.errno != errno.EEXIST:
                raise

    def _getfname(self, outname):
        return path.join(self.jugdir, outname[0], outname[1], outname[2:])


    def dump(self, object, outname):
        '''
        dump(outname, object)

        Performs the same as

        pickle.dump(object, file(outname,'w')

        but does it in a way that is guaranteed to be atomic even over NFS.
        '''
        # Rename is atomic in NFS, but we need to create the temporary file in
        # the same directory as the result.
        #
        # Don't mess unless you know what you are doing!
        outname = self._getfname(outname)

        # Format:
        #   If extension is .npy.gz, then it's a gzipped numpy npy file
        #   If extension is .pp.gz, then it's a gzipped python pickle
        #   If extension is .pp, then it's a python pickle
        #   If extension is empty and file is empty, then it's None
        if object is None:
            create_directories(dirname(outname))
            F=file(outname,'w')
            F.close()
            return
        extension = '.pp.gz'
        write = pickle.dump
        if isinstance(object, np.ndarray):
            extension = '.npy.gz'
            write = (lambda f,a: np.save(a,f))
        outname = outname + extension
        fd, fname = tempfile.mkstemp(extension,'jugtemp',options.tempdir)
        os.close(fd)
        F = GzipFile(fname,'w')
        write(object,F)
        F.close()
        create_directories(dirname(outname))
        os.rename(fname,outname)


    def can_load(self, fname):
        '''
        can = can_load(fname)
        '''
        fname = self._getfname(fname)
        return exists(fname + '.pp.gz') or exists(fname) or exists(fname + '.npy.gz')

    def load(self, fname):
        '''
        obj = load(fname)

        Loads the objects. Equivalent to pickle.load(), but a bit smarter at times.
        '''
        fname = self._getfname(fname)
        if exists(fname) and _fsize(fname) == 0:
            return None
        elif exists(fname + '.pp.gz'):
            return pickle.load(GzipFile(fname + '.pp.gz'))
        elif exists(fname + '.npy.gz'):
            return np.load(GzipFile(fname + '.npy.gz'))
        else:
            return pickle.load(file(fname))

    def remove(self, fname):
        '''
        was_removed = remove(fname)

        Remove the entry associated with fname.

        Returns whether any entry was actually removed.
        '''
        fname = self._getfname(fname)
        possible = [fname,fname+'.pp.gz',fname+'.npy.gz']
        for p in possible:
            if os.path.exists(p):
                os.unlink(p)
                return True
        return False

    def cleanup(self, active):
        '''
        cleanup()

        Implement 'cleanup' command
        '''
        files = set()
        for path,_,fs in os.walk(self.jugdir):
            for f in fs:
                files.add(path+'/'+f)
        for t in tasks:
            fname = t.hash()
            possible = [fname,fname+'.pp.gz',fname+'.npy.gz']
            for p in possible:
                try:
                    files.remove(p)
                except:
                    pass
        for f in files:
            os.unlink(f)
        print 'Removed %s files' % len(files)

    def getlock(self, name):
        return file_based_lock(name)

    def close(self):
        '''
        store.close()

        Has no effect on file based stores.
        '''
        pass


def _fullname(name):
    return path.join(options.lockdir,name+'.lock')

class file_based_lock(object):
    '''
    file_based_lock: File-system based locks

    Functions:
    ----------

        * get(): acquire the lock
        * release(): release the lock
        * is_locked(): check lock state
    '''

    def __init__(self, name):
        self.fullname = _fullname(name)

    def get(self):
        '''
        lock.get()

        Create a lock for name in an NFS compatible way.
        '''
        if exists(self.fullname): return False
        create_directories(path.dirname(self.fullname))
        try:
            fd = os.open(self.fullname,os.O_RDWR|os.O_CREAT|os.O_EXCL)
            F = os.fdopen(fd,'w')
            print >>F, 'Lock', os.getpid()
            F.close()
            return True
        except OSError:
            return False

    def release(self):
        '''
        lock.release()

        Removes lock
        '''
        try:
            os.unlink(self.fullname)
        except OSError:
            pass

    def is_locked(self):
        '''
        locked = lock.is_locked()

        Returns whether a lock exists for name. Note that the answer can
        be invalid by the time this function returns. Only by trying to
        acquire the lock can you avoid race-conditions. See the get() function.
        '''
        return path.exists(self.fullname)

# vim: set ts=4 sts=4 sw=4 expandtab smartindent:



# -*- coding: utf-8 -*-
# Copyright (C) 2008-2009, Luís Pedro Coelho <lpc@cmu.edu>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
'''
Task: contains the Task class.

This is the main unit of jug.
'''

from __future__ import division
import hashlib
import os
import cPickle as pickle

alltasks = []

class Task(object):
    '''
    Task
    ----

    T = Task(f, dependencies, fwkwargs)

    Defines a task, which is roughly equivalent to

    f( *[dep() for dep in dependencies], **fkwargs)

    '''
    store = None
    def __init__(self,f,*dependencies, **kwdependencies):
        if f.func_name == '<lambda>':
            raise ValueError('''jug.Task does not work with lambda functions!

Write an email to the authors if you feel you have a strong reason to use them (they are a bit
tricky to support since the general code relies on the function name)''')

        kwargs = dict( (k,v) for k,v in kwdependencies.iteritems() if k.startswith('task_') )
        for k in kwargs:
            del kwdependencies[k]

        self.name = '%s.%s' % (f.__module__, f.__name__)
        self.f = f
        self.dependencies = dependencies
        self.kwdependencies = kwdependencies
        self.finished = False
        self.loaded = False
        self._hash = None
        self._lock = None
        self._can_load = False
        self.print_result = kwargs.get('task_print_result',False)
        alltasks.append(self)

    def run(self, force=False):
        '''
        task.run(force=False)

        Performs the task.

        Parameters
        ----------
          * force: if True, always run the task (even if it ran before)
        '''
        assert self.can_run()
        assert force or not self.finished
        args = [value(dep) for dep in self.dependencies]
        kwargs = dict((key,value(dep)) for key,dep in self.kwdependencies.iteritems())
        self._result = self.f(*args,**kwargs)
        name = self.hash()
        self.store.dump(self._result, name)
        self.finished = True
        if self.print_result:
            print self._result
        return self._result

    def _get_result(self):
        if not self.finished: self.load()
        return self._result

    result = property(_get_result,doc='Result value')


    def can_run(self):
        '''
        bool = task.can_run()

        Returns true if all the dependencies are finished.
        '''
        def is_available(dep):
            if type(dep) == Task: return dep.finished or dep.can_load()
            if type(dep) == list: return all(is_available(sub) for sub in dep)
            return True # If dependency is not list nor task, it's a literal value
        return all(is_available(dep) for dep in (list(self.dependencies) + self.kwdependencies.values()))

    def load(self):
        '''
        t.load()

        Loads the results from file.
        '''
        assert self.can_load()
        self._result = self.store.load(self.hash())
        self.finished = True

    def unload(self):
        '''
        t.unload()

        Unload results (can be useful for saving memory)
        '''
        self._result = None
        self.finished = False

    def unload_recursive(self):
        '''
        t.unload_recursive()

        '''
        def _dependency_walk(task):
            for dep in list(task.dependencies) + task.kwdependencies.values():
                if type(dep) == Task:
                    for d in _dependency_walk(dep):
                        yield d
                yield dep
        for dep in _dependency_walk(self):
            if type(dep) == Task:
                dep.unload()
        self.unload()

    def can_load(self):
        '''
        bool = task.can_load()
        '''
        if self.finished: return True
        if not self._can_load:
            self._can_load = self.store.can_load(self.hash())
        return self._can_load

    def hash(self):
        '''
        fname = t.hash()

        Returns the hash for this task
        '''
        if self._hash is None:
            M = hashlib.md5()
            def update(*args):
                if not args: return
                names,elems = args
                for n,e in zip(names,elems):
                    M.update(pickle.dumps(n))
                    if type(e) == Task:
                        M.update(e.hash())
                    elif type(e) == list:
                        update(*zip(*enumerate(e)))
                    elif type(e) == dict:
                        update(e.keys(),e.values())
                    else:
                        M.update(pickle.dumps(e))
            M.update(self.name)
            update(*zip(*enumerate(self.dependencies)))
            update(*zip(*self.kwdependencies.items()))
            M.update(pickle.dumps(self.name))
            self._hash = M.hexdigest()
        return self._hash

    def __str__(self):
        '''String representation'''
        return 'Task: %s()' % self.name

    def __repr__(self):
        '''Detailed representation'''
        return 'Task(%s,dependencies=%s,kwdependencies=%s)' % (self.name,self.dependencies,self.kwdependencies)

    def lock(self):
        '''
        locked = task.lock()

        Tries to lock the task for the current process.
        Returns true if the lock was obtained.
        '''
        if self._lock is None:
            self._lock = self.store.getlock(self.hash())
        return self._lock.get()

    def unlock(self):
        '''
        task.unlock()

        Releases the lock.
        If the lock was not held, this may remove another
        thread's lock!
        '''
        self._lock.release()

    def is_locked(self):
        if self._lock is None:
            self._lock = self.store.getlock(self.hash())
        return self._lock.is_locked()


def topological_sort(tasks):
    '''
    topological_sort(tasks)

    Sorts a list of tasks topologically in-place. The list is sorted when
    there is never a dependency between tasks[i] and tasks[j] if i < j.
    '''
    sorted = []
    whites = set(tasks)
    def dfs(t):
        for dep in recursive_dependencies(t, 1):
            if dep in whites:
                whites.remove(dep)
                dfs(dep)
        sorted.append(t)
    while whites:
        next = whites.pop()
        dfs(next)
    tasks[:] = sorted

def recursive_dependencies(t, max_level=-1):
    '''
    for dep in recursive_dependencies(t, max_level=None):
        ...

    Returns a generator that lists all recursive dependencies of task

    Parameters
    ----------
        * t: task
        * max_level: Maximum recursion depth. Set to -1 or None for no recursion limit.
    '''
    if max_level is None:
        max_level = -1
    if type(t) is list:
        for d in t:
            for dd in recursive_dependencies(d, max_level):
                yield dd
        return
    if type(t) is dict:
        for d in t.itervalues():
            for dd in recursive_dependencies(d, max_level):
                yield dd
        return
    if type(t) is Task:
        yield t
        if max_level == 0:
            return
        for dep in t.dependencies:
            for d in recursive_dependencies(dep, max_level-1):
                yield d
        for dep in t.kwdependencies.itervalues():
            for d in recursive_dependencies(dep, max_level-1):
                yield d


def value(elem):
    '''
    value = value(task)

    Loads a task object recursively.
    '''
    if type(elem) is Task:
        return elem.result
    elif type(elem) is list:
        return [value(e) for e in elem]
    elif type(elem) is tuple:
        return tuple([value(e) for e in elem])
    elif type(elem) is dict:
        return dict((x,value(y)) for x,y in elem.iteritems())
    else:
        return elem

def CachedFunction(f,*args,**kwargs):
    '''
    value = CachedFunction(f, *args, **kwargs)

    is the same as

    value = Task(f, *args, **kwargs).result

    That is, it calls the function if the value is available,
    but caches the result for the future.
    '''
    t = Task(f,*args, **kwargs)
    if not t.can_load():
        if not t.can_run():
            raise ValueError('jug.CachedFunction: unable to run task %s' % t)
        t.run()
    return t.result

def TaskGenerator(*args,**kwargs):
    '''
    TaskGenerator

    Use as either

    @TaskGenerator
    def f():
        pass

    or

    @TaskGenerator(arg=1)
    def f():
        pass

    TaskGenerator does not take any non-keyword arguments
    '''
    assert not (args and kwargs), '''TaskGenerator called in a weird way.'''
    if kwargs:
        task_args = [('task_' + k,v) for k,v in kwargs.iteritems()]
        def task_generator(f):
            def gen_task(*args,**kwargs):
                kwargs.update(task_args)
                return Task(f,*args,**kwargs)
            return gen_task
        return task_generator
    assert len(args) == 1, '''TaskGenerator called in a weird way.'''
    func = args[0]
    def gen_task(*args,**kwargs):
        return Task(func,*args,**kwargs)
    return gen_task

# vim: set ts=4 sts=4 sw=4 expandtab smartindent:



# -*- coding: utf-8 -*-
# Copyright (C) 2008-2009, Luís Pedro Coelho <lpc@cmu.edu>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.
'''
Options

Variables
---------

    * jugdir: main jug directory.
    * tempdir: directory for temporary files.
    * lockdir: directory for lock files.
    * jugfile: filesystem name for the Jugfile
    * cmd: command to run.
    * aggressive_unload: --aggressive-unload
    * invalid_name: --invalid
    * other_args: Arguments not captured by jug (for script use)
'''
from __future__ import division

jugdir = 'jugdata'
tempdir = jugdir + '/tempfiles/'
lockdir = jugdir + '/locks/'
jugfile = 'jugfile.py'
cmd = None
aggressive_unload = False
invalid_name = None
other_args = None

_Commands = ('execute','status','stats','cleanup','count','invalidate')
_Usage_string = \
'''python %s COMMAND OPTIONS...

Possible commands:
* execute
    Execute tasks
* status:
    Print status
* counts:
    Simply count tasks
* cleanup:
    Cleanup
* stats
    Print statistics [Not implemented]
* invalidate
    Invalidate the results of a task

Options:
--jugfile=JUGFILE
    Name of the jugfile to use (if not given, use jugfile.py)
--aggressive-unload
    Aggressively unload data from memory
--invalid=TASK-NAME
    Task name to invalidate (for invalidate command only)
--jugdir=JUGDIR
    Directory in which to save intermediate files
'''

def usage():
    '''
    usage()

    Print an usage string and exit.
    '''
    import sys
    print _Usage_string % sys.argv[0]
    sys.exit(1)

def find_jugfile(options):
    '''
    jugfile = find_jugfile(options)

    Returns the filename of the Jugfile or None if it is not found.
    '''
    if options.jugfile:
        return options.jugfile
    return 'jugfile.py'

def parse():
    '''
    options.parse()

    Parse the command line options and set the option variables.
    '''
    import optparse
    global cmd, jugfile, aggressive_unload, invalid_name, other_args
    global jugdir, tempdir, lockdir
    parser = optparse.OptionParser()
    parser.add_option('--jugfile',action='store',type='string',dest='jugfile',default=None)
    parser.add_option('--aggressive-unload',action='store_true',dest='aggressive_unload',default=False)
    parser.add_option('--invalid',action='store',dest='invalid_name',default=None)
    parser.add_option('--jugdir',action='store',dest='jugdir',default='jugdata/')
    options,args = parser.parse_args()
    if not args:
        usage()
        return
    cmd = args[0]
    if cmd not in _Commands:
        usage()
        return
    if options.invalid_name and cmd != 'invalidate':
        usage()
        return
    if cmd == 'invalidate' and not options.invalid_name:
        usage()
        return

    aggressive_unload = options.aggressive_unload
    jugfile = find_jugfile(options)
    invalid_name = options.invalid_name
    other_args = args[1:]
    jugdir = options.jugdir
    tempdir = jugdir + '/tempfiles/'
    lockdir = jugdir + '/locks/'

# vim: set ts=4 sts=4 sw=4 expandtab smartindent:



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2008-2009, Luís Pedro Coelho <lpc@cmu.edu>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from __future__ import division
from collections import defaultdict
from signal import signal, SIGTERM
from time import sleep
import sys
import os
import os.path
import random
import logging

from . import options
from . import task
from . import file_based_store
from . import redis_store
from .task import Task

silent = False
def print_out(s):
    if not silent:
        print s

def do_print(store):
    '''
    do_print(store)

    Print a count of task names.
    '''
    task_counts = defaultdict(int)
    for t in task.alltasks:
        task_counts[t.name] += 1
    for tnc in task_counts.items():
        print 'Task %s: %s' % tnc

def invalidate(store):
    '''
    invalidate()

    Implement 'invalidate' command
    '''
    invalid_name = options.invalid_name
    invalid = set()
    tasks = task.alltasks
    for t in tasks:
        if t.name == invalid_name:
            invalid.add(t)
        else:
            for dep in task.recursive_dependencies(t):
                if type(dep) is task.Task:
                    if dep.name == invalid_name:
                        invalid.add(t)
                        break
    if not invalid:
        print 'No results invalidated.'
        return
    task_counts = defaultdict(int)
    for t in invalid:
        if store.remove(t.hash()):
            task_counts[t.name] += 1
    if sum(task_counts.values()) == 0:
        print 'Tasks invalidated, but no results removed'
    else:
        print 'Tasks Invalidated'
        print
        print '    Task Name          Count'
        print '----------------------------'
        for n_c in task_counts.items():
            print '%21s: %12s' % n_c


def execute(store):
    '''
    execute(store)

    Implement 'execute' command
    '''
    task_names = set(t.name for t in task.alltasks)
    tasks = task.alltasks
    tasks_executed = defaultdict(int)
    tasks_loaded = defaultdict(int)
    task.topological_sort(tasks)
    logging.info('Execute start (%s tasks)' % len(tasks))
    signal(SIGTERM,_sigterm)
    waits = [0,4,8,16,32,64,128,128,128,128,1024,2048]
    upnext = []
    while tasks or upnext:
        if not upnext:
            for w in waits:
                if w:
                    logging.info('waiting...', w, 'for an open task')
                    sleep(w)
                upnext = []
                i = 0
                while i < len(tasks):
                    if tasks[i].can_run():
                        upnext.append(tasks[i])
                        del tasks[i]
                        if len(upnext) > 10:
                            break
                    else:
                        i += 1
                if upnext:
                    break
            if not upnext:
                logging.info('No tasks can be run!')
                return
        t = upnext.pop(0)
        if t.can_load():
            logging.info('Loadable %s...' % t.name)
            tasks_loaded[t.name] += 1
            continue
        locked = False
        try:
            locked = t.lock()
            if t.can_load(): # This can be true if the task run between the check above and this one
                logging.info('Loadable %s...' % t.name)
                tasks_loaded[t.name] += 1
            elif locked:
                logging.info('Executing %s...' % t.name)
                t.run()
                tasks_executed[t.name] += 1
                if options.aggressive_unload:
                    t.unload_recursive()
            else:
                logging.info('Already in execution %s...' % t.name)
        except Exception, e:
            print >>sys.stderr, 'Exception while running %s: %s' % (t.name,e)
            raise
        finally:
            if locked: t.unlock()

    print_out('%-52s%12s%12s' %('Task name','Executed','Loaded'))
    print_out('-' * (52+12+12))
    for t in task_names:
        print_out('%-52s%12s%12s' % (t,tasks_executed[t],tasks_loaded[t]))
    if not task_names:
        print_out('<no tasks>')

def status(store):
    '''
    status(store)

    Implements the status command.
    '''
    task_names = set(t.name for t in task.alltasks)
    tasks = task.alltasks
    tasks_ready = defaultdict(int)
    tasks_finished = defaultdict(int)
    tasks_running = defaultdict(int)
    tasks_waiting = defaultdict(int)
    for t in tasks:
        if t.can_load():
            tasks_finished[t.name] += 1
        elif t.can_run():
            if t.is_locked():
                tasks_running[t.name] += 1
            else:
                tasks_ready[t.name] += 1
        else:
            tasks_waiting[t.name] += 1

    print '%-40s%12s%12s%12s%12s' %('Task name','Waiting','Ready','Finished','Running')
    print ('-' * (40+12+12+12+12))
    for t in task_names:
        print '%-40s%12s%12s%12s%12s' % (t[:40],tasks_waiting[t],tasks_ready[t],tasks_finished[t],tasks_running[t])
    print ('.' * (40+12+12+12+12))
    print '%-40s%12s%12s%12s%12s' % ('Total:',sum(tasks_waiting.values()),sum(tasks_ready.values()),sum(tasks_finished.values()),sum(tasks_running.values()))
    print

def cleanup(store):
    '''
    cleanup(store)

    Implement 'cleanup' command
    '''
    tasks = task.alltasks
    removed = store.cleanup(tasks)
    print 'Removed %s files' % removed

def init():
    '''
    init()

    Initializes jug (creates needed directories &c).
    Imports jugfile
    '''
    jugdir = options.jugdir
    if jugdir.startswith('redis:'):
        store = redis_store.redis_store(options.jugdir)
    else:
        store = file_based_store.file_store(options.jugdir)
    Task.store = store

    jugfile = options.jugfile
    if jugfile.endswith('.py'):
        jugfile = jugfile[:-len('.py')]
    sys.path.insert(0, os.path.abspath('.'))
    try:
        __import__(jugfile)
    except ImportError:
        print >>sys.stderr, "Could not import file '%s'" % jugfile
        sys.exit(1)
    return store


def _sigterm(_,__):
    sys.exit(1)

def main():
    options.parse()
    store = init()

    if options.cmd == 'execute':
        execute(store)
    elif options.cmd == 'count':
        do_print(store)
    elif options.cmd == 'status':
        status(store)
    elif options.cmd == 'invalidate':
        invalidate(store)
    elif options.cmd == 'cleanup':
        cleanup(store)
    else:
        print >>sys.stderr, 'Jug: unknown command: \'%s\'' % options.cmd

if __name__ == '__main__':
    try:
        main()
    except Exception, exc:
        print >>sys.stderr, 'Jug Error!'
        raise
# vim: set ts=4 sts=4 sw=4 expandtab smartindent:



from jug.file_based_store import file_based_lock

def test_twice():
    lock = file_based_lock('foo')
    assert lock.get()
    assert not lock.get()
    lock.release()

    assert lock.get()
    assert not lock.get()
    lock.release()

def test_twolocks():
    foo = file_based_lock('foo')
    bar = file_based_lock('bar')
    assert foo.get()
    assert bar.get()
    assert not foo.get()
    assert not bar.get()
    foo.release()
    bar.release()




# -*- coding: utf-8 -*-
# Copyright (C) 2008-2009, Luís Pedro Coelho <lpc@cmu.edu>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#  THE SOFTWARE.

from __future__ import division
from sys import exit
try:
    import setuptools
except:
    print '''
setuptools not found. Please install it.

On linux, the package is often called python-setuptools'''
    exit(1)

long_description = '''\
Jug: A Task-Based Parallelization Framework
-------------------------------------------

Jug allows you to write code that is broken up into
tasks and run different tasks on different processors.

It has two storage backends: One uses the filesystem to
communicate between processes and works correctly over NFS,
so you can coordinate processes on different machines. The
other uses a redis database and all it needs is for different
processes to be able to communicate with a common redis server.

Jug is a pure Python implementation and should work on any platform.
'''

classifiers = [
'Development Status :: 3 - Alpha',
'Environment :: Console',
'License :: OSI Approved :: MIT License',
'Operating System :: POSIX',
'Operating System :: OS Independent',
'Programming Language :: Python',
'Topic :: Scientific/Engineering',
'Topic :: Software Development',
]

setuptools.setup(name = 'Jug',
      version = '0.5.0-beta-0',
      description = 'A Task Based Parallelization Framework',
      long_description = long_description,
      author = 'Luis Pedro Coelho',
      author_email = 'lpc@mcu.edu',
      license = 'MIT',
      platforms = ['Any'],
      classifiers = classifiers,
      url = 'http://luispedro.org/jug',
      packages = setuptools.find_packages(exclude='tests'),
      scripts = ['bin/jug'],
      test_suite = 'nose.collector',
      )


# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
